/*-------------------------------------+-------------------------------------*/
/*                                 control.c                                 */

#include <iostream.h>
#include "hardbord.h"
#include "softbord.h"

enum cmdtype {
   NOP, HELP, GERMAN, ENGLISH, FRENCH, BOARD, FEN, LAN, SAN,
   HINT, DEPTH, MOVE, AUTO, DEMO, TOUR,
   OOO, OO, STRING, OTHER, UNDO, START, END, POS, NEW, QUIT
};

static struct {
   const char * name;
   cmdtype command;
} cmdlist[] = {
   { "",        NOP     },
   { "?",       HELP    },
   { "help",    HELP    },
   { "german",  GERMAN  },
   { "english", ENGLISH },
   { "french",  FRENCH  },
   { ".",       BOARD   },
   { "board",   BOARD   },
   { "fen",     FEN     },
   { "lan",     LAN     },
   { "san",     SAN     },

   { "hint",    HINT    },
   { "depth",   DEPTH   },
   { "move",    MOVE    },
   { "auto",    AUTO    },
   { "demo",    DEMO    },
   { "tour",    TOUR    },

   { "o-o-o",   OOO     },
   { "0-0-0",   OOO     },
   { "OOO",     OOO     },
   { "ooo",     OOO     },
   { "000",     OOO     },
   { "o-o",     OO      },
   { "0-0",     OO      },
   { "OO",      OO      },
   { "oo",      OO      },
   { "00",      OO      },

   { "other",   OTHER   },
   { "undo",    UNDO    },
   { "{",       START   },
   { "}",       END     },
   { "pos",     POS     },
   { "new",     NEW     },
   { "quit",    QUIT    }
};

static void show_help (void)
{
   cout << endl <<
   "?, help          display this help screen                      " << endl <<
   "german           display and accept German letters for pieces  " << endl <<
   "english          display and accept English letters for pieces " << endl <<
   "french           display and accept French letters for pieces  " << endl <<
   "., board         display an ASCII board for this position      " << endl <<
   "fen              show the FEN notation for this position       " << endl <<
   "lan              show the LAN list of possible moves           " << endl <<
   "san              show the SAN list of possible moves           " << endl <<
   "hint <f> <n>     let function <f> give a tree search evaluation" << endl <<
   "depth <n>        set search depth, default is 2                " << endl <<
   "move <f>         let function <f> make a move                  " << endl <<
   "auto <f>,off     let function <f> play against you, default Eva" << endl <<
   "demo <f>         let function <f> finish this variation        " << endl <<
   "tour <f> <f> <n> let the functions make an n-game tournament   " << endl <<
   "other            undo the last half move                       " << endl <<
   "undo             undo the last full move                       " << endl <<
   "{                (start a variation  (can be nested))          " << endl <<
   "}                delete the innermost variation                " << endl <<
   "pos <string>     setup a position from FEN notation            " << endl <<
   "new              new game                                      " << endl <<
   "quit             quit this program, like EOF                   " << endl <<
   endl;
   return;
}

static cmdtype lookup (const char * s)
{
   unsigned i;
   for (i=0; i < sizeof(cmdlist)/sizeof(*cmdlist); i++) {
      unsigned j;
      for (j=0; s[j] && s[j]==cmdlist[i].name[j]; j++)
         ;
      if (s[j] == cmdlist[i].name[j])
         return cmdlist[i].command;
   }
   return STRING;
}

const unsigned linelen = 100;
static char cs[linelen];

static cmdtype getcmd (void)
{
   if (cin.eof ())
      return QUIT;
   else {
      for (unsigned i=0; i<6; i++)
         cs[i] = '\0';
      cin >> cs;
      return lookup (cs);
   }
}


extern struct {
   const char * name;
   long (*pointer)(void);
} efnlist[];

static inline char low (char c)
{
   return ('A'<=c && c<='Z') ? c+'a'-'A' : c;
}

long (* name (const char * s)) (void)
{
   unsigned i;
   for (i=0; efnlist[i].name; i++) {
      unsigned j;
      for (j=0; s[j] && s[j]==efnlist[i].name[j]; j++)
         ;
      if (s[j]==efnlist[i].name[j] || s[j]==low(efnlist[i].name[j]))
         return efnlist[i].pointer;
   }
   cout << "Function " << s << " not implemented." << endl;
   cout << "Here's my list of available evaluation functions:" << endl;
   for (i=0; efnlist[i].name; i++)
      cout << "\t" << efnlist[i].name << endl;
   return NULL;
}


void control (void)
{
   const char * language = "PNBRQK";
   hardboard * g = new hardboard;
   g->show_board ();
   unsigned n;
   unsigned depth = 2;
   long (*autoanswer)(void) = name ("Eva");

   while (TRUE)
      switch (getcmd ()) {
      case NOP:
         break;
      case HELP:
         show_help ();
         break;
      case GERMAN:
         language = "OSLTDK";
         g->setlang (language);
         g->show_board ();
         break;
      case ENGLISH:
         language = "PNBRQK";
         g->setlang (language);
         g->show_board ();
         break;
      case FRENCH:
         language = "PCFTDR";
         g->setlang (language);
         g->show_board ();
         break;
      case BOARD:
         g->show_board ();
         break;
      case FEN:
         g->show_fen ();
         break;
      case LAN:
         g->show_lan ();
         break;
      case SAN:
         g->show_san ();
         break;

      case HINT:
         cin >> cs >> n;
         if (name (cs)) {
            softboard v (name (cs));
            if (v.treesearch (n)) {
               cout << cs << ": " <<
                  g->expand (v.result.p0, v.result.f1, v.result.promotion) <<
                  " is good (" << v.result.value << " millipawn)";
               if (v.result.value > big/2)
                  cout << ", is a mate in " << (big-v.result.value+1)/2;
               else if (v.result.value < -big/2)
                  cout << ", is a mate in " << (big+v.result.value+1)/2;
               cout << "." << endl;
            } else if (! n)
               cout << cs << ": " << "This position is worth " <<
                  v.result.value << " millipawns." << endl;
            else if (! v.result.value)
               cout << "There is no move.  You're stalemate." << endl;
            else
               cout << "There is no move.  You've been mated." << endl;
         }
         break;
      case DEPTH:
         cin >> depth;
         if (! depth)
            depth++;
         cout << "Tree search depth set to " << depth << "." << endl;
         break;
      case MOVE:
         cin >> cs;
         if (name (cs)) {
            softboard v (name (cs));
            if (v.treesearch (depth)) {
               cout << g->expand (v.result.p0, v.result.f1, v.result.promotion)
                  << endl;
               g->execute (v.result.p0, v.result.f1, v.result.promotion);
               g->show_board ();
            } else if (! v.result.value)
               cout << "There is no move.  It's a stalemate." << endl;
            else
               cout << "There is no escape.  It's a mate." << endl;
         }
         break;
      case AUTO:
         cin >> cs;
         if (cs[0]=='o' && cs[1]=='f' && cs[2]=='f' && cs[3]=='\0')
            autoanswer = NULL;
         else if (name (cs))
            autoanswer = name (cs);
         break;
      case DEMO:
         cin >> cs;
         if (name (cs)) {
            softboard v (name (cs));
            while (g->movelist && ! g->drawn_by_3 () && ! g->drawn_by_50 ()) {
               v.treesearch (depth);
               cout << g->expand (v.result.p0, v.result.f1, v.result.promotion)
                  << endl;
               g->execute (v.result.p0, v.result.f1, v.result.promotion);
               g->show_board ();
            }
            if (! g->movelist && attacked (piecelist[color][4].loc))
               cout << (color ? "Black" : "White") << " is mate." << endl;
            else if (! g->movelist)
               cout << (color ? "Black" : "White") << " is stalemate." << endl;
            else if (g->drawn_by_3 ())
               cout << "Drawn by third-repetition rule." << endl;
            else
               cout << "Drawn by 50 move rule." << endl;
         }
         break;
      case TOUR:
         {
         char f[2][linelen];
         unsigned rounds;
         cin >> f[0] >> f[1] >> rounds;
         if (name (f[0]) && name (f[1])) {
            unsigned ismate[2]={0,0}, isstale[2]={0,0}, draw3=0, draw50=0;
            while (rounds--) {
               delete g;
               g = new hardboard;
               g->setlang (language);
               unsigned who = rounds%2==!color;
               softboard *v = new softboard (name (f[who]));
               cout << endl << "Game " << f[who] << " (white) against " <<
                  f[!who] << " (black):" << endl;
               g->show_board ();
               while (g->movelist && ! g->drawn_by_3() && ! g->drawn_by_50()) {
                  v->treesearch (depth);
                  cout << g->expand (v->result.p0, v->result.f1,
                     v->result.promotion) << endl;
                  g->execute (v->result.p0, v->result.f1, v->result.promotion);
                  g->show_board ();
                  delete v;
                  who = rounds%2==!color;
                  v = new softboard (name (f[who]));
               }
               if (! g->movelist && attacked (piecelist[color][4].loc)) {
                  ismate[who]++;
                  cout << (color ? "Black" : "White") <<
                     " (" << f[who] << ") is mate." << endl;
               } else if (! g->movelist) {
                  isstale[who]++;
                  cout << (color ? "Black" : "White") <<
                     " (" << f[who] << ") is stalemate." << endl;
               } else if (g->drawn_by_3 ()){
                  draw3++;
                  cout << "Drawn by third-repetition rule." << endl;
               } else {
                  draw50++;
                  cout << "Drawn by 50 move rule." << endl;
               }
               delete v;
            }
            cout << endl <<
               "Result:" << endl <<
               f[0]<<     " mated "<<f[1]<<" "<< ismate[1]<<" times."<< endl <<
               f[0]<<" stalemated "<<f[1]<<" "<<isstale[1]<<" times."<< endl <<
               "Drawn by third-repetition rule "<< draw3  <<" times."<< endl <<
               "Drawn by 50 move rule "         << draw50 <<" times."<< endl <<
               f[1]<<" stalemated "<<f[0]<<" "<<isstale[0]<<" times."<< endl <<
               f[1]<<     " mated "<<f[0]<<" "<< ismate[0]<<" times."<< endl <<
               endl;
         }
         }
         break;

      case OOO:
         cs[0]='O', cs[1]='-', cs[2]='O', cs[3]='-', cs[4]='O';
      case OO:
         cs[0]='O', cs[1]='-', cs[2]='O';
      case STRING:
         n = g->execute (cs);
         if (n == 1) {
            g->show_board ();
            if (autoanswer) {
               softboard v (autoanswer);
               if (v.treesearch (depth)) {
                  cout << g->expand (v.result.p0, v.result.f1,
                     v.result.promotion) << endl;
                  g->execute (v.result.p0, v.result.f1, v.result.promotion);
                  g->show_board ();
               } else if (! v.result.value)
                  cout << "There is no move.  I'm stalemate." << endl;
               else
                  cout << "There is no escape.  I've been mated." << endl;
            }
         } else {
            if (n)
               cout << "Ambiguous input matches " << n << " moves." << endl;
            else
               cout << "No such move." << endl;
            cout << "Possible moves are:" << endl;
            g->show_san ();
         }
         break;
      case OTHER:
         g->undo ();
         g->show_board ();
         break;
      case UNDO:
         g->undo ();
         g->undo ();
         g->show_board ();
         break;
      case START:
         g->start_variation ();
         break;
      case END:
         g->end_variation ();
         g->show_board ();
         break;
      case POS:
         {
            unsigned i = 0;
            for (n=0; n<6; n++) {
               cin >> (cs+i);
               while (cs[i])
                  i++;
               cs[i++] = ' ';
            }
            cs[i-1] = '\0';
         }
         delete g;
         g = new hardboard (cs, language);
         g->setlang (language);
         cout << "pos ";
         g->show_fen ();
         g->show_board ();
         break;
      case NEW:
         delete g;
         g = new hardboard;
         g->setlang (language);
         g->show_board ();
         break;
      case QUIT:
         delete g;
         return;
      }
}

